from dataclasses import dataclass, field
from enum import Enum

from typing import List, Dict


class TestMessageLevel(Enum):
    ERROR = (1,)
    INFO = 2


@dataclass
class TestMessage:
    type: TestMessageLevel
    text: str

    def _repr_html_(self):
        return """
               <li style="color:{0};">{1}</li>
               """.format('red' if self.type == TestMessageLevel.ERROR else 'inherit', self.text)

    def __repr__(self):
        return f" - [{self.type}] {self.text}"


@dataclass
class PartialUnexpectedCounts:
    value: List[int]
    count: int


@dataclass
class TestResult:
    """
    Dataclass representing the result of a test

    :param passed: A boolean indicating whether the test passed or not
    :param messages: A list of TestMessage objects containing information about the test execution
    :param metric: A float representing the test metric
    """
    passed: bool = False
    messages: List[TestMessage] = field(default_factory=list, repr=False)
    props: Dict[str, str] = field(default_factory=dict, repr=False)
    metric: float = 0
    missing_count: int = 0
    missing_percent: float = 0
    unexpected_count: int = 0
    unexpected_percent: float = 0
    unexpected_percent_total: float = 0
    unexpected_percent_nonmissing: float = 0
    partial_unexpected_index_list: List[PartialUnexpectedCounts] = field(default_factory=list, repr=False)
    unexpected_index_list: List[int] = field(default_factory=list, repr=False)
    output_df: bytes = None
    number_of_perturbed_rows: int = 0
    actual_slices_size: List[int] = field(default_factory=list, repr=False)
    reference_slices_size: List[int] = field(default_factory=list, repr=False)

    def _repr_html_(self):
        return """
               <h4><span style="color:{0};">{1}</span> Test {2}</h4>
               <p>Metric: {3}<p>
               <ul>{4}</ul>
               """.format('green' if self.passed else 'red',
                          '✓' if self.passed else '𐄂',
                          'succeed' if self.passed else 'failed',
                          'No metric' if self.metric is None else str(round(self.metric, 2)),
                          ''.join([] if self.messages is None else [m._repr_html_() for m in self.messages]))

    def __repr__(self):
        return """
               Test {0}
               Metric: {1}
               {2}
               """.format('succeed' if self.passed else 'failed',
                          'No metric' if self.metric is None else str(round(self.metric, 2)),
                          '\n'.join([] if self.messages is None else [m.__repr__() for m in self.messages]))



import pandas as pd
import pytest

import giskard.testing.tests.metamorphic as metamorphic
from giskard.datasets.base import Dataset
from giskard.ml_worker.testing.registry.transformation_function import transformation_function


def _test_metamorphic_increasing_regression(ds: Dataset, model, threshold):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.bmi = x.bmi + x.bmi * 0.1
        return x

    results = metamorphic.test_metamorphic_increasing(model=model, dataset=ds, transformation_function=perturbation,
                                                      threshold=threshold).execute()

    assert results.actual_slices_size[0] == 442
    assert round(results.metric, 2) == 0.44
    return results.passed


def _test_metamorphic_decreasing_regression(ds: Dataset, model, threshold):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.age = x.age - x.age * 0.1
        return x

    results = metamorphic.test_metamorphic_decreasing(model=model, dataset=ds, transformation_function=perturbation,
                                                      threshold=threshold).execute()
    assert results.actual_slices_size[0] == 442
    assert round(results.metric, 2) == 0.54
    return results.passed


def _test_metamorphic_increasing_classification(df, model, threshold):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.duration_in_month = x.duration_in_month + x.duration_in_month * 0.5
        return x

    results = metamorphic.test_metamorphic_increasing(model=model, dataset=df, transformation_function=perturbation,
                                                      threshold=threshold,
                                                      classification_label=model.meta.classification_labels[
                                                          0]).execute()

    assert results.actual_slices_size[0] == 1000
    assert results.metric == 1
    return results.passed


def _test_metamorphic_decreasing_classification(df, model, threshold):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.duration_in_month = x.duration_in_month * 0.5
        return x

    results = metamorphic.test_metamorphic_decreasing(model=model, dataset=df, transformation_function=perturbation,
                                                      threshold=threshold,
                                                      classification_label=model.meta.classification_labels[
                                                          0]).execute()

    assert results.actual_slices_size[0] == 1000
    assert results.metric == 1
    return results.passed


def test_metamorphic_increasing_classification(german_credit_test_data, german_credit_model):
    assert _test_metamorphic_increasing_classification(german_credit_test_data, german_credit_model, 0.8)


def test_metamorphic_decreasing_classification(german_credit_test_data, german_credit_model):
    assert _test_metamorphic_decreasing_classification(german_credit_test_data, german_credit_model, 0.8)


def test_metamorphic_increasing_regression(diabetes_dataset, linear_regression_diabetes):
    assert _test_metamorphic_increasing_regression(diabetes_dataset, linear_regression_diabetes, 0.3)
    assert not _test_metamorphic_increasing_regression(diabetes_dataset, linear_regression_diabetes, 0.5)


def test_metamorphic_decreasing_regression(diabetes_dataset, linear_regression_diabetes):
    assert _test_metamorphic_decreasing_regression(diabetes_dataset, linear_regression_diabetes, 0.5)
    assert not _test_metamorphic_decreasing_regression(diabetes_dataset, linear_regression_diabetes, 0.6)


def test_metamorphic_decreasing_exception(german_credit_test_data, german_credit_model):
    with pytest.raises(Exception):
        @transformation_function()
        def perturbation(x: pd.Series) -> pd.Series:
            x.duration_in_month = x.duration_in_month * 0.5
            return x

        metamorphic.test_metamorphic_decreasing(model=german_credit_model, dataset=german_credit_test_data,
                                                transformation_function=perturbation, threshold=0.5,
                                                classification_label="random_value").execute()


def test_metamorphic_increasing_exception(german_credit_test_data, german_credit_model):
    with pytest.raises(Exception):
        @transformation_function()
        def perturbation(x: pd.Series) -> pd.Series:
            x.duration_in_month = x.duration_in_month * 0.5
            return x

        metamorphic.test_metamorphic_increasing(model=german_credit_model, dataset=german_credit_test_data,
                                                transformation_function=perturbation, threshold=0.5,
                                                classification_label="random_value").execute()


def test_metamorphic_increasing_t_test(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.sex = "female" if x.sex == "male" else "male"
        return x

    results = metamorphic.test_metamorphic_increasing_t_test(model=german_credit_model, dataset=german_credit_test_data,
                                                             transformation_function=perturbation,
                                                             critical_quantile=0.05,
                                                             classification_label="Default").execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert not results.passed, f"metric = {results.metric}"


def test_metamorphic_decreasing_t_test(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.sex = "female" if x.sex == "male" else "male"
        return x

    results = metamorphic.test_metamorphic_decreasing_t_test(model=german_credit_model, dataset=german_credit_test_data,
                                                             transformation_function=perturbation,
                                                             critical_quantile=0.05,
                                                             classification_label="Default").execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert results.passed, f"metric = {results.metric}"


def test_metamorphic_increasing_wilcoxon(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.sex = "female" if x.sex == "male" else "male"
        return x

    results = metamorphic.test_metamorphic_increasing_wilcoxon(model=german_credit_model,
                                                               dataset=german_credit_test_data,
                                                               transformation_function=perturbation,
                                                               critical_quantile=0.05,
                                                               classification_label="Default").execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert not results.passed, f"metric = {results.metric}"


def test_metamorphic_decreasing_wilcoxon(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        x.sex = "female" if x.sex == "male" else "male"
        return x

    results = metamorphic.test_metamorphic_decreasing_wilcoxon(model=german_credit_model,
                                                               dataset=german_credit_test_data,
                                                               transformation_function=perturbation,
                                                               critical_quantile=0.05,
                                                               classification_label="Default").execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert results.passed, f"metric = {results.metric}"


def test_metamorphic_increasing_t_test_nopert(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        return x

    results = metamorphic.test_metamorphic_increasing_t_test(model=german_credit_model, dataset=german_credit_test_data,
                                                             transformation_function=perturbation,
                                                             critical_quantile=0.05,
                                                             classification_label="Default").execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert not results.passed, f"metric = {results.metric}"


def test_metamorphic_decreasing_t_test_nopert(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        return x

    results = metamorphic.test_metamorphic_decreasing_t_test(model=german_credit_model, dataset=german_credit_test_data,
                                                             transformation_function=perturbation,
                                                             critical_quantile=0.05,
                                                             classification_label="Default").execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert not results.passed, f"metric = {results.metric}"


def test_metamorphic_increasing_wilcoxon_nopert(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        return x

    results = metamorphic.test_metamorphic_increasing_wilcoxon(model=german_credit_model,
                                                               dataset=german_credit_test_data,
                                                               transformation_function=perturbation,
                                                               critical_quantile=0.05,
                                                               classification_label="Default").execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert not results.passed, f"metric = {results.metric}"


def test_metamorphic_decreasing_wilcoxon_nopert(german_credit_test_data, german_credit_model):
    @transformation_function()
    def perturbation(x: pd.Series) -> pd.Series:
        return x

    results = metamorphic.test_metamorphic_decreasing_wilcoxon(model=german_credit_model,
                                                               dataset=german_credit_test_data,
                                                               transformation_function=perturbation,
                                                               critical_quantile=0.05,
                                                               classification_label="Default").execute()

    assert results.actual_slices_size[0] == len(german_credit_test_data)
    assert not results.passed, f"metric = {results.metric}"



import codecs
import datetime
import logging
import os
import re
import socket
from contextlib import closing

import requests

import giskard
from giskard.client.python_utils import warning
from giskard.utils import threaded

logger = logging.getLogger(__name__)


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def readable_hex(data):
    if not os.environ.get("GSK_ML_WORKER_LOG_HEX", False):
        return ""
    s = codecs.encode(data, "hex").decode()
    return " ".join(s[i: i + 2] for i in range(0, len(s), 2))


def is_pre_release(current_version: str):
    # A very dummy pre-release checker, to be improved
    return re.match(r'.*[a-z].*', current_version) is not None


@threaded
def check_latest_giskard_version():
    try:
        current_version = giskard.__version__
        if not is_pre_release(current_version):
            return
        respose = requests.get('https://pypi.org/pypi/giskard/json').json()
        releases = respose.get('releases')
        releases_dates = {}
        if current_version not in releases:
            return
        for ver, resources in releases.items():
            latest_release_date = max(map(lambda r: datetime.datetime.fromisoformat(r['upload_time']), resources))
            releases_dates[ver] = latest_release_date
        latest_version, latest_release_date = max(releases_dates.items(), key=lambda x: x[1])

        if latest_release_date > releases_dates[current_version]:
            warning(
                "You're using a pre-release version of giskard while a "
                "new version is available, please install it with: "
                f"pip install \"giskard=={latest_version}\""
            )
    except BaseException as e:
        logger.exception("Failed to fetch latest giskard version", e)



from typing import Sequence

from .base_detector import BaseTextPerturbationDetector
from .text_transformations import TextTransformation
from ...datasets.base import Dataset
from ...models.base import BaseModel
from ..decorators import detector


@detector(
    name="text_perturbation",
    tags=["text_perturbation", "robustness", "nlp", "classification", "regression", "generative", "text_generation", "llm"],
)
class TextPerturbationDetector(BaseTextPerturbationDetector):
    def _get_default_transformations(self, model: BaseModel, dataset: Dataset) -> Sequence[TextTransformation]:
        from .text_transformations import (
            TextUppercase,
            TextLowercase,
            TextTitleCase,
            TextTypoTransformation,
            TextPunctuationRemovalTransformation,
        )

        return [
            TextUppercase,
            TextLowercase,
            TextTitleCase,
            TextTypoTransformation,
            TextPunctuationRemovalTransformation,
        ]
